<?php

namespace Drupal\Tests\whitelabel\Kernel;

use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\user\Traits\UserCreationTrait;
use Drupal\Tests\whitelabel\Traits\WhiteLabelCreationTrait;
use Drupal\whitelabel\PathProcessor\WhiteLabelPathProcessor;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Tests if the correct style sheets are applied to a page.
 *
 * @group whitelabel
 */
class WhiteLabelInboundPathProcessingTest extends KernelTestBase {

  use WhiteLabelCreationTrait {
    createWhiteLabel as drupalCreateWhiteLabel;
  }
  use UserCreationTrait {
    createUser as drupalCreateUser;
  }

  /**
   * {@inheritdoc}
   */
  public static $modules = [
    'system',
    'text',
    'options',
    'user',
    'file',
    'image',
    'whitelabel',
    'node',
  ];

  /**
   * Holds the generated white label throughout the different tests.
   *
   * @var \Drupal\whitelabel\WhiteLabelInterface
   */
  private $whiteLabel;

  /**
   * Holds the randomly generated token.
   *
   * @var string
   */
  private $token;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->installSchema('system', ['sequences']);
    $this->installConfig(['whitelabel']);
    $this->installEntitySchema('whitelabel');
    $this->installEntitySchema('user');

    $user = $this->drupalCreateUser(['serve white label pages']);
    $this->setCurrentUser($user);

    $this->token = $this->randomMachineName(16);
    $this->whiteLabel = $this->createWhiteLabel(['token' => $this->token]);
    $this->whiteLabelProvider = \Drupal::service('whitelabel.whitelabel_provider');
  }

  /**
   * Test to see if the module returns white labeled pages for all modes.
   * @dataProvider provideModes
   */
  public function testInboundPaths($mode) {
    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());

    // Configure the white label mode.
    $this->config('whitelabel.settings')->set('mode', $mode)->save();

    list ($path, $request) = $this->generateRequest($mode);
    $this->processInbound($path, $request);

    // Ensure White Label equals the expected one.
    $this->assertEquals($this->whiteLabel->id(), $this->whiteLabelProvider->getWhiteLabel()->id());
  }

  /**
   * Test to see if non existing tokens still resolve pages.
   * @dataProvider provideModes
   */
  public function testFakeToken($mode) {
    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());

    // Configure the white label mode.
    $this->config('whitelabel.settings')->set('mode', $mode)->save();

    $expect_404 = FALSE;
    if ($mode == WhiteLabelPathProcessor::CONFIG_PATH_PREFIX || $mode == WhiteLabelPathProcessor::CONFIG_DOMAIN) {
      $expect_404 = TRUE;
    }

    list ($path, $request) = $this->generateRequest($mode, 'someFakeToken');
    $this->processInbound($path, $request, $expect_404);

    // Ensure White Label is still empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());
  }

  /**
   * Test to see if tokens based on system paths are correctly resolved.
   * @dataProvider provideModes
   */
  public function testTrickyTokens($mode) {
    // TODO: Enable node module to test that as well.

    $tricky_tokens = [
      'admin',
      'node',
      'system',
      'user',
    ];

    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());

    // Configure the white label mode.
    $this->config('whitelabel.settings')->set('mode', $mode)->save();

    foreach ($tricky_tokens as $tricky_token) {
      $this->whiteLabel->setToken($tricky_token)->save();

      foreach ($tricky_tokens as $tricky_path) {
        list ($path, $request) = $this->generateRequest($mode, $tricky_token, $tricky_path);
        $this->processInbound($path, $request);

        $this->assertEquals($this->whiteLabel->id(), $this->whiteLabelProvider->getWhiteLabel()->id(), 'Asserting token: ' . $tricky_token .  ' for path: ' . $tricky_path);
      }
    }
  }

  /**
   * Test to see if less privileged user gets a no WL page.
   * @dataProvider provideModes
   */
  public function testViewerNoPermissions($mode) {
    // Test with unprivileged user.
    $white_label_viewer = $this->drupalCreateUser();
    $this->setCurrentUser($white_label_viewer);

    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());

    // Configure the white label mode.
    $this->config('whitelabel.settings')->set('mode', $mode)->save();

    list ($path, $request) = $this->generateRequest($mode);
    $this->processInbound($path, $request);

    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());
  }

  /**
   * Test to see if less privileged owner serves a no WL page.
   * @dataProvider provideModes
   */
  public function testOwnerNoPermissions($mode) {
    // Test with regular viewer account.
    $white_label_viewer = $this->drupalCreateUser(['view white label pages']);
    $this->setCurrentUser($white_label_viewer);

    // Change the owner of the white label to a less privileged one.
    $white_label_owner = $this->drupalCreateUser();
    $this->whiteLabel->setOwner($white_label_owner)->save();

    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());

    // Configure the white label mode.
    $this->config('whitelabel.settings')->set('mode', $mode)->save();

    $expect_404 = FALSE;
    if ($mode == WhiteLabelPathProcessor::CONFIG_PATH_PREFIX || $mode == WhiteLabelPathProcessor::CONFIG_DOMAIN) {
      $expect_404 = TRUE;
    }

    list ($path, $request) = $this->generateRequest($mode);
    $this->processInbound($path, $request, $expect_404);

    // Ensure White Label is empty.
    $this->assertNull($this->whiteLabelProvider->getWhiteLabel());
  }

  /**
   * @return array
   *   Return an array with White label modes
   */
  public function provideModes() {
    return [
      [
        WhiteLabelPathProcessor::CONFIG_QUERY_PARAMETER,
      ],
      [
        WhiteLabelPathProcessor::CONFIG_PATH_PREFIX,
      ],
      [
        WhiteLabelPathProcessor::CONFIG_DOMAIN,
      ],
    ];
  }

  /**
   * Constructs an incoming request for the given mode and token.
   *
   * @param string $mode
   *   The white label mode to return the request for.
   * @param string|null $token
   *   The token to use when constructing the request, or NULL to use default.
   *
   * @return array
   *   An array with the uri and the Symfony\Component\HttpFoundation\Request.
   */
  protected function generateRequest($mode, $token = NULL, $base_path = '') {
    if (empty($token)) {
      $token = $this->token;
    }

    $uri = $base_path;
    if ($mode === WhiteLabelPathProcessor::CONFIG_PATH_PREFIX) {
      $uri .= '/' . $token;
    }

    $requests = [
      WhiteLabelPathProcessor::CONFIG_QUERY_PARAMETER => Request::create($uri, 'GET', ['whitelabel' => $token], [], [], []),
      WhiteLabelPathProcessor::CONFIG_PATH_PREFIX => Request::create($uri, 'GET', [], [], [], []),
      WhiteLabelPathProcessor::CONFIG_DOMAIN => Request::create($uri, 'GET', [], [], [], ['HTTP_HOST' => $token . '.localhost']),
    ];

    return [$uri, $requests[$mode]];
  }

  /**
   * Processes the inbound request.
   *
   * Under normal circumstances, this should set the white label as defined in
   * the inbound path processor. After calling this, we can assert that the
   * white label has indeed been set.
   *
   * @see \Drupal\whitelabel\PathProcessor\WhiteLabelPathProcessor
   *
   * @param string $path
   *   The path to process.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The constructed request to test with.
   * @param bool $expect_404
   *   Boolean indicating if a NotFoundHttpException is expected.
   *
   * @return null|string
   *   Returns the path provided by the path processors, or NULL on faillure.
   * @throws \Exception
   */
  protected function processInbound($path, Request $request, $expect_404 = FALSE) {
    try {
      $new_path = $this->container->get('path_processor_manager')->processInbound($path, $request);
      return $new_path;
    }
    catch (NotFoundHttpException $exception) {
      if (!$expect_404) {
        throw new \Exception('NotFoundHttpException was thrown, but this exception was not expected. Exception: ' . $exception->getMessage());
      }
      // Exception was expected, so do nothing.
    }
    catch (\Exception $exception) {
      throw new \Exception('Unexpected exception encountered');
    }

    return NULL;
  }

}
